import {
  ConflictException,
  Injectable,
  NotFoundException,
} from '@nestjs/common';
import { InjectRepository } from '@nestjs/typeorm';
import { FindOptionsWhere, Repository } from 'typeorm';
import { InstitucionCarrera } from '../institucion-carrera/entity/institucion-carrera.entity';
import { InstitucionUsuario } from '../institucion-usuario/entity/institucion-usuario.entity';
import { Operador } from '../operador/entity/operador.entity';
import { TipoUsuario } from '../tipo-usuario/entity/tipo-usuario.entity';
import { Usuario } from './entity/usuario.entity';
import { FullInformacionUsuarioView } from './entity/views/full-informacion-usuario.view';
import { BcryptService } from '../bcrypt/bcrypt.service';
import { InstitucionService } from '../institucion/institucion.service';
import { InstitucionCarreraService } from '../institucion-carrera/institucion-carrera.service';
import { InstitucionUsuarioService } from '../institucion-usuario/institucion-usuario.service';
import { NodemailerService } from '../nodemailer/nodemailer.service';
import { TipoUsuarioService } from '../tipo-usuario/tipo-usuario.service';

@Injectable()
export class UsuarioService {
  constructor(
    @InjectRepository(Usuario) private repository: Repository<Usuario>,
    @InjectRepository(FullInformacionUsuarioView)
    private fullInformacionUsuarioView: Repository<FullInformacionUsuarioView>,
    private bcryptService: BcryptService,
    private institucionService: InstitucionService,
    private institucionCarreraService: InstitucionCarreraService,
    private institucionUsuarioService: InstitucionUsuarioService,
    private nodemailerService: NodemailerService,
    private tipoUsuarioService: TipoUsuarioService,
  ) {}

  create(
    usuario: string,
    nombre: string,
    tipoUsuario: TipoUsuario,
    institucionCarrera: InstitucionCarrera,
    correo: string,
    rfc?: string,
  ): Promise<InstitucionUsuario> {
    // Creamos y guardamos un registro
    return this.repository
      .save(
        this.repository.create({
          nombre,
          tipoUsuario,
          usuario,
          correo,
          rfc: rfc || '',
        }),
      )
      .then((usuario) =>
        // Le asignamos la carrera enviada
        this.institucionUsuarioService.create(institucionCarrera, usuario),
      );
  }

  async findAll(filtros: {
    pagina: string;
    activo?: string;
    id_carrera?: string;
    id_institucion?: string;
    id_tipo_usuario?: string;
    nombre?: string;
    password?: string;
    usuario?: string;
  }): Promise<[Usuario[], number]> {
    const carrera = filtros.id_carrera
      ? await this.institucionCarreraService.findCarreraByIdCarrera(
          parseInt(filtros.id_carrera),
        )
      : null;
    const institucion = filtros.id_institucion
      ? await this.institucionService.findById(parseInt(filtros.id_institucion))
      : null;
    const tipoUsuario = filtros.id_tipo_usuario
      ? await this.tipoUsuarioService.findById(
          parseInt(filtros.id_tipo_usuario),
        )
      : null;
    const query = this.repository
      .createQueryBuilder('u')
      .innerJoinAndSelect('u.instituciones', 'is')
      .innerJoinAndSelect('u.tipoUsuario', 'tu')
      .innerJoinAndSelect('is.institucionCarrera', 'ic')
      .innerJoinAndSelect('ic.carrera', 'c')
      .innerJoinAndSelect('ic.institucion', 'i')
      .innerJoinAndSelect('c.nivel', 'n')
      .take(25)
      .skip((parseInt(filtros.pagina) - 1) * 25);

    if (
      filtros.password &&
      (filtros.password === 'true' || filtros.password === 'false')
    )
      query.andWhere(
        `u.password IS ${filtros.password === 'true' ? 'NOT' : ''} NULL`,
      );
    if (filtros.activo)
      query.andWhere(`is.activo = :activo`, {
        activo: filtros.activo === 'true',
      });
    if (filtros.usuario)
      query.andWhere('u.usuario LIKE :usuario', {
        usuario: `%${filtros.usuario}%`,
      });
    if (filtros.nombre)
      query.andWhere('u.nombre LIKE :nombre', {
        nombre: `%${filtros.nombre}%`,
      });
    if (carrera)
      query.andWhere('c.id_carrera = :id_carrera', {
        id_carrera: carrera.id_carrera,
      });
    if (institucion)
      query.andWhere('i.id_institucion = :id_institucion', {
        id_institucion: institucion.id_institucion,
      });
    if (tipoUsuario)
      query.andWhere('tu.id_tipo_usuario = :id_tipo_usuario', {
        id_tipo_usuario: tipoUsuario.id_tipo_usuario,
      });
    return query.getManyAndCount();
  }

  findFullInfoByUsuario(operador: Operador, usuario: string) {
    return this.fullInformacionUsuarioView
      .findOne({ where: { usuario } })
      .then(async (infoUsuario) => {
        if (!infoUsuario)
          throw new NotFoundException('No existe este usuario.');

        const usuario = this.repository.create({
          id_usuario: infoUsuario.id_usuario,
          correo: infoUsuario.correo,
          telefono: infoUsuario.telefono,
          nombre: infoUsuario.nombre,
          rfc: infoUsuario.rfc,
          usuario: infoUsuario.usuario,
          tipoUsuario: {
            id_tipo_usuario: infoUsuario.id_tipo_usuario,
            tipo_usuario: infoUsuario.tipo_usuario,
          },
        });

        usuario.instituciones =
          await this.institucionUsuarioService.findFullInfoAllByIdUsuarioIdInstitucion(
            usuario,
            operador.institucion,
          );
        return usuario;
      });
  }

  findById(id_usuario: number) {
    return this.repository
      .findOne({
        select: ['id_usuario', 'correo', 'password', 'usuario'],
        where: { id_usuario },
      })
      .then((usuario) => {
        if (!usuario) throw new NotFoundException('No existe este usuario.');
        return usuario;
      });
  }

  findByUsuario(usuario: string, rfc?: string) {
    const busqueda: FindOptionsWhere<Usuario> = { usuario };

    if (rfc) busqueda.rfc = rfc;
    return this.repository.findOne({
      select: ['id_usuario', 'correo', 'nombre', 'password', 'usuario'],
      where: busqueda,
    });
  }

  passwordReset(
    operador: Operador,
    id_usuario: number,
  ): Promise<{ message: string }> {
    const password = this.bcryptService.generarPassword(); // Generamos una password

    return this.findById(id_usuario)
      .then(async (usuario) => {
        await this.validarsOperadorMod(usuario, operador);
        // Solo se puede reenviar password si no esta registrad@, tener password = registro
        if (!usuario.password)
          throw new ConflictException(
            'No se puede enviar una contraseña a un usuario no registrado.',
          );
        // Actualizamos password
        usuario.password = this.bcryptService.encriptar(password);
        // Guardamos cambios
        return this.repository.save(usuario);
      })
      .then((usuario) =>
        // Enviamos un correo con credenciales
        this.nodemailerService.sendEmail({
          email: usuario.correo,
          subject: 'Credenciales Pc Puma',
          html: this.nodemailerService.correoPasswordUsuario(password),
        }),
      )
      .then((_) => ({
        message: 'Se cambió correctamente la constraseña del usuario.',
      }));
  }

  async registrar(
    id_institucion_carrera: number,
    usuario: string,
    telefono: string,
    rfc = '',
  ): Promise<{ message: string }> {
    const institucionCarrera = await this.institucionCarreraService.findById(
      id_institucion_carrera,
    );
    const password = this.bcryptService.generarPassword(); // Generamos un password
    let message: string;

    return this.findByUsuario(usuario, rfc)
      .then(async (existeUsuario) => {
        // Si no existe o tiene contraseña, es decir que esta registrado, sacamos error
        if (!existeUsuario || existeUsuario.password)
          throw new ConflictException(
            'Verificar sus datos, cualquier duda asistir al módulo PC PUMA.',
          );
        // Validamos que el alunmo pertenezca a la carrera enviada
        return this.institucionUsuarioService.findFullInfoByIdUsuarioIdInstitucionCarrera(
          existeUsuario,
          institucionCarrera,
        );
      })
      .then((institucionUsuario) => {
        // Si no tiene correo saca error
        if (!institucionUsuario.usuario.correo)
          throw new ConflictException(
            'Tu institución no te asignó un correo institucional, dirigete al módulo para solucionarlo.',
          );

        // variable para guardar la posición del @
        const arrobaIndex = institucionUsuario.usuario.correo.indexOf('@');

        // Mensaje personalizado, blureando el correo al que se va a enviar las credenciales
        message = `Se creó correctamente tu cuenta, mandamos un correo a ...${institucionUsuario.usuario.correo.slice(
          arrobaIndex - 3,
        )} con tus credenciales para acceder a este servicio.`;
        // Asignar datos enviados del front y encriptar la password
        institucionUsuario.usuario.telefono = telefono;
        institucionUsuario.usuario.password =
          this.bcryptService.encriptar(password);
        // Guardamos cambios
        return this.repository.save(institucionUsuario.usuario);
      })
      .then((usuario) =>
        // Enviamos un correo con credenciales
        this.nodemailerService.sendEmail({
          email: usuario.correo,
          subject: 'Credenciales Pc Puma',
          html: this.nodemailerService.correoPasswordUsuario(password),
        }),
      )
      .then((_) => ({ message }));
  }

  save(usuario: Usuario): Promise<{ message: string }> {
    return this.repository
      .save(usuario)
      .then((_) => ({ message: 'Se guardaron los cambios correctamente.' }));
  }

  update(
    operador: Operador,
    attrs: Partial<Usuario>,
  ): Promise<{ message: string }> {
    return this.findById(attrs.id_usuario)
      .then(async (usuario) => {
        await this.validarsOperadorMod(usuario, operador);
        // Asignamos valores enviados al objeto
        Object.assign(usuario, attrs);
        // Guardamos cambios
        return this.repository.save(usuario);
      })
      .then((_) => ({ message: 'Se guardaron los cambios correctamente.' }));
  }

  private async validarsOperadorMod(
    usuario: Usuario,
    operador: Operador,
  ): Promise<void> {
    return this.institucionUsuarioService
      .findAllByIdUsuarioIdInstitucion(usuario, operador.institucion)
      .then((institucionesUsuario) => {
        // Si el array no trae elementos significa que el usuario existe pero no le pertenece a la institución de admin
        if (institucionesUsuario.length === 0)
          throw new ConflictException(
            'No puedes modificar la información este usuario porque no pertenece a tu institución.',
          );
      });
  }
}
